#!/bin/sh
#
# ScratchBox toolchain floating point test script.
#
# 

CC=gcc
CFLAGS=-g
LDFLAGS=-g
HEADERS="stdio.h"
SIGNALS=""
LIBS=""

# HEADERS="stdio.h math.h"
# LIBS="-lm"


##
# Usage: c_expr <expression>...
#	expression:	test expression
c_expr() {
	equ=no
	cmd="c_expr $@"
	var=""
	vars=""
	stmt=""
	count=0
	tmpfile="c_expr"

	while test $# != 0; do
		case $1 in
			var|static)
				# If variable is a float and initialized with a immediate, append
				# 'f' to the initializer to make sure compiler handles it as a
				# float.
				if expr $2 == float \& $3 : '[0-9]*\.[0-9]\+$' &>/dev/null; then
					var="$2 var${count} = $3f;"
				else
					var="$2 var${count} = $3;"
				fi
				
				if test $1 = "static"; then
					var="static ${var}"
				fi

				vars="${vars}\n\t${var}"
				stmt="${stmt} var${count}"

				var=""
				count=`expr ${count} + 1`
				tmpfile="${tmpfile}_$1_$2_$3"
				shift 3
				;;

			cast)
				stmt="${stmt} (($2) $3)"
				tmpfile="${tmpfile}_$1_$2_$3"
				shift 3
				;;

			==|!=|\<|\>|\<=|=\>)
				equ=yes
				stmt="${stmt}) == ("
				tmpfile="${tmpfile}_equ"
				shift
				;;

			+)
				tmpfile="${tmpfile}_plus"
				stmt="${stmt} $1"
				shift
				;;

			-)
				tmpfile="${tmpfile}_minus"
				stmt="${stmt} $1"
				shift
				;;

			%)
				tmpfile="${tmpfile}_mod"
				stmt="${stmt} $1"
				shift
				;;

			\*)
				tmpfile="${tmpfile}_mul"
				stmt="${stmt} $1"
				shift
				;;

			/)
				tmpfile="${tmpfile}_div"
				stmt="${stmt} $1"
				shift
				;;

			*)
				tmpfile="${tmpfile}_$1"
				stmt="${stmt} $1"
				shift
				;;
		esac

	done

	tmpfile="${tmpfile}"
	srcfile="${tmpfile}.c"
	# srcfile=/dev/stdout
	objfile="${tmpfile}.o"
	exefile="${tmpfile}.exe"

	# FIXME:
	rm -f "${srcfile}" "${objfile}" "${exefile}"

	if test -f "${srcfile}" -o -f "${objfile}" -o -f "${exefile}"; then
		echo "Warning! Skipping test because temporary file already exists. " >&2
		echo "Remove all files created by previous test runs before running " >&2
		echo "float test." >&2
		return 1
	fi

	datetime=`date '+%A %d.%m.%Y %X'`
	echo "/* This program was automatically created ${datetime} " >> "${srcfile}"
	echo " * by ScratchBox compiler toolchain floating point test suite." >> "${srcfile}"
	echo " * " >> "${srcfile}"
	echo " * Command used to create this file was:" >> "${srcfile}"
	echo " *     ${cmd}" >> "${srcfile}"
	echo " */" >> "${srcfile}"

	for header in ${HEADERS}; do
		echo "#include <${header}>" >> "${srcfile}"
	done

	for signal in ${SIGNALS}; do
		echo -e "\nvoid sighandler_${signal}(int signum) {" >> "${srcfile}"
		eval "echo \"	\${SIGHANDLER_${signal}}\"" >> "${srcfile}"
		echo "}" >> "${srcfile}"
	done

	echo -ne "\nint main(void) {" >> "${srcfile}"
	echo -ne "${vars}\n" >> "${srcfile}"

	for signal in ${SIGNALS}; do
		echo -e "\tsignal(${signal}, sighandler_${signal});" >> "${srcfile}"
	done

	if test ${equ} = yes; then
		echo -ne "\treturn(((${stmt})) ? 0 : -1);\n" >> "${srcfile}"
	else
		echo -ne "\treturn(${stmt});\n" >> "${srcfile}"
	fi
	echo -ne "}\n" >> "${srcfile}"


#	return 0

	echo -n "Compiling ${srcfile}... "
	${CC} ${CFLAGS} -o "${objfile}" -c "${srcfile}"
	retval=$?
	if test "x${retval}" != "x0"; then
		echo "failed!"
		return ${retval}
	fi

	echo "ok."
	sync
	sleep 1

	echo -n "Linking ${objfile}.. "	
	${CC} ${LDFLAGS} -o "${exefile}" ${LIBS} "${objfile}"
	retval=$?
	if test "x${retval}" != "x0"; then
		echo "failed!"
		return ${retval}
	fi
	
	echo "ok."
	sync

	echo -n "Running ${exefile}... "
	./"${exefile}"
	retval=$?
	if test "x${retval}" != "x0"; then
		echo "failed!"
		return ${retval}
	fi
	
	echo "ok."

	# FIXME:
	rm -f "${exefile}" "${srcfile}" "${objfile}"
	
	return 0
}

##
# Usage: c_trap <signal> <code>...
#	signal	signal to catch
#	code	code to run, when signal is caught
c_trap() {
	# 
	if test $# -lt 2; then
		echo "function c_trap() requires atleast two parameters."
		return 1
	fi
	
	# First argument is name of signal. Let's make sure it's in uppercase
	# like it's in C header files.
	signal="`echo $1 | tr 'a-z' 'A-Z'`"
	shift
	
	# Everything else is copied in to handler function body.
	# FIXME: '-characters must be escaped.
	body="$@"

	# Update list of trapped signals...
	SIGNALS="${SIGNALS} ${signal}"

	# Make sure signal.h is included.
	if ! expr "${HEADERS}" : '.*signal\.h' &>/dev/null ; then
		HEADERS="${HEADERS} signal.h"
	fi
	
	eval "export SIGHANDLER_${signal}='$@'"
}

##
# Usage: c_untrap <signal>
#	signal	signal to uncatch
c_untrap() {
	# Let's make sure signal name is in uppercase like it's in C header
	# files.
	signal="`echo $1 | tr 'a-z' 'A-Z'`"

	SIGNALS="`echo "${SIGNALS}" | sed -e 's/[ ]*$1//`"
	
	eval "unset SIGHANDLER_${signal}"
}


##
# Usage: test_float_op <type1> <value1> <operation> <type2> <value2> <result>
#	type1:		type of first operand
#	value1:		value of first operand
#	operation:	type of operation
#	type2:		type of second operand
#	value2:		value of second operand
#	result:		result
test_float_op() {
	IFS=$' \t\n'

	t1=$1
	v1=$2
	op=$3
	t2=$4
	v2=$5
	rs=$6
	for q1 in static var cast; do
		for q2 in static var cast; do
			c_expr "${q1}" "${t1}" "${v1}" "${op}" "${q2}" "${t2}" "${v2}" == ${rs}
		done
	done
}

types='float:double:long double:char:int:long long'

c_trap SIGFPE 'fprintf(stderr, "Floating point exception!\n");' exit\(-2\)\;
c_trap SIGILL 'fprintf(stderr, "Illegal instruction!\n");' exit\(-3\)\;
c_trap SIGSEGV 'fprintf(stderr, "Segmentation fault!\n");' exit\(-4\)\;
c_trap SIGTERM 'fprintf(stderr, "SIGTERM received!\n");' exit\(-5\)\;


TFS="${IFS}"
IFS=':'
for ot1 in ${types}; do
	IFS=':'
	for ot2 in ${types}; do
		IFS="${TFS}"
		echo "--- ${ot1} --- ${ot2} ---"
		test_float_op "${ot1}" 2.0 \+ "${ot2}" 4.0 6.0
		test_float_op "${ot1}" 6.0 \- "${ot2}" 4.0 2.0
		test_float_op "${ot1}" 6.0 \/ "${ot2}" 3.0 2.0
		test_float_op "${ot1}" 2.0 \* "${ot2}" 3.0 6.0
	done
done
